#
# $Copyright (c) 2021 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas Technologies
# LLC or its affiliates in the U.S. and other countries. Other names may be trademarks
# of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $
#

# Max msg id = 47

eval 'exec /opt/VRTSperl/bin/perl -I `pwd`/../../lib -S $0 ${1+"$@"}'
        if 0;

package GoldenGateAgent;
use ag_i18n_inc;
use ArrayRep;
@ISA = ("ArrayRep");

use lib qw(.);
use ggconfig () ;
use vars qw(%CONF);
*CONF = \%ggconfig::CONF;
my $SU = $CONF{SU};
my $PASSWDFILE = $CONF{PASSWDFILE};
my $LD_ENV;

%default = {
	resname =>'',
	ora_res =>'',
	extract =>'',
	data_pump =>'',
	replicat =>'',
	auto_takeover =>'',
	goldengate_home =>'',
	owner =>'',

	sid =>'',
	oracle_home=>'',
	goldengate_bin =>'',
	user =>'',
	domain =>'',
	password =>'',
};

sub new {
        my $proto = shift;
        my $resname = shift;
        my @attrs = @_;
        my $class = ref($proto) || $proto;
        my $self = { %default };
        $self->{'resname'} = $resname;
        bless($self, $class);
        my $ret = $VCSAG_SUCCESS;
        $self->{'start_time'} = $^T;
        $self->{'offline_timeout'} = 0;

	($ret, $self->{'ora_res'}) = VCSAG_GET_ATTR_VALUE ("LinkRes", -1, 1, @attrs);
        exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'extract'}) = VCSAG_GET_ATTR_VALUE ("Extract", -1, 1, @attrs);
        exit 0 if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'replicat'}) = VCSAG_GET_ATTR_VALUE ("Replicat", -1, 1, @attrs);
        exit 0 if ($ret != $VCSAG_SUCCESS);
	
	($ret, $self->{'data_pump'}) = VCSAG_GET_ATTR_VALUE ("DataPump", -1, 1, @attrs);
	
	($ret, $self->{'auto_takeover'}) = VCSAG_GET_ATTR_VALUE ("AutoTakeover", -1, 1, @attrs);
        exit 0 if ($ret != $VCSAG_SUCCESS);
	
	($ret, $self->{'goldengate_home'}) = VCSAG_GET_ATTR_VALUE ("GoldenGateHome", -1, 1, @attrs);
        exit 0 if ($ret != $VCSAG_SUCCESS);
	
	return $self;
}


#validates agent arguments
#RC = 0, validation succeeded
#RC = 1, validation failed
sub validate_args {
        my $self = shift;
        my $ret = 0;
 
        if ($self->{'ora_res'} eq "") {
                VCSAG_LOG_MSG("E", "LinkRes must be specified.", 1);
                $ret = 1;
	} 
       	if ($self->{'extract'} eq "") {
                VCSAG_LOG_MSG("E", "Extract must be specified.", 2);
                $ret = 1;
        }
       	if ($self->{'replicat'} eq "") {
                VCSAG_LOG_MSG("E", "Replicat must be specified.", 3);
                $ret = 1;
        }
       	if ($self->{'goldengate_home'} eq "") {
                VCSAG_LOG_MSG("E", "GoldenGateHome must be specified.", 4);
               $ret = 1;
        } else {
		$ret = $self->check_goldengate_home($self->{'goldengate_home'});
	}
	if (($self->{'auto_takeover'} != 0) && ($self->{'auto_takeover'} != 1)) {
                VCSAG_LOG_MSG("E", "AutoTakeover must have value of 0 or 1.", 5);
                $ret = 1;
        }
	return $ret;
}

sub initialize {
	my $self = shift;
	my $ret = 0;
	if ($^O =~ m/MSWin32/i ) {
		my ($Sid, $User, $Domain, $Password, $Home);
	} else {
		my ($Sid, $Owner, $Home, $EnvFile);
	}
	my $ValidFile = 0;

	my $LocSys = $self->get_local_nodename();
	VCSAG_LOGDBG_MSG(1, "Got local nodename :: $LocSys", 6, "$LocSys");
	#$LocSys = "WIN-GG-VM3";
	$self->{'local_system'} = $LocSys;
	
	if ($^O =~ m/MSWin32/i ) {
		$ret = $self->get_link_res_attr($self->{'ora_res'}, $LocSys, $Sid, $User, $Domain, $Password);
	} else {
		$ret = $self->get_link_res_attr($self->{'ora_res'}, $LocSys, $Sid, $Owner, $Home, $EnvFile);
	}
	
	if ($ret) {
        	VCSAG_LOG_MSG ("E","Oracle Link resource $LinkRes does not exist on node $LocSys", 7, $LinkRes, $LocSys);
        	return 1;
	}

	if ($^O =~ m/MSWin32/i ) {
		VCSAG_LOGDBG_MSG(1, "Oracle link resource SID = $Sid,  Oracle Home = $Home, User = $User, Domain = $Domain", 47, "$Sid", "$Home", "$User", "$Domain");
		$self->{'user'} = $User;
		$self->{'domain'} = $Domain;
		$self->{'password'} = $Password;
	} else {
		VCSAG_LOGDBG_MSG(1, "Oracle link resource SID = $Sid, Owner = $Owner, Oracle Home = $Home, EnvFile = $EnvFile", 8, "$Sid", "$Owner", "$Home", "$EnvFile");
		$self->{'owner'} = $Owner;
		$self->{'env_file'} = $EnvFile;

		$LD_ENV = $self->set_instance_env($EnvFile, $Owner, $Home, $ValidFile);
		VCSAG_LOGDBG_MSG(1, "LD_ENV :: $LD_ENV", 9, "LD_ENV");
		$self->{'oracle_home'} = $Home;
	
		$ENV{'ORACLE_HOME'} = $Home;
	}
		
	$self->{'sid'} = $Sid;
	$ENV{'ORACLE_SID'} = $Sid;	

	return $ret;
}

# Function : check_goldengate_home
# Desc     : Checks for inexistence of GoldenGate Home
# Input    : $Home 
# Output   : None
# Returns  : (1) -  home does not exist  
#
sub check_goldengate_home  {
	my $self = shift;
	my ($Home) =@_;
	my $gg_bin = "";
	unless ( -d $Home ) {
		VCSAG_LOG_MSG("E", "GoldenGateHome does not exist.", 10);
		return  1;
	}
	if ($^O =~ m/MSWin32/i ) {
		$self->{'gg_bin'} = "$Home\\ggsci.exe";
	} else {
		 $self->{'gg_bin'} = "$Home/ggsci";
	}
	unless ( -x  $self->{'gg_bin'}) {
		VCSAG_LOG_MSG("E", "Oracle GoldenGate Command Interpreter does not exist in GoldenGateHome $Home.", 11, "$Home");
		return 1;
	}
	return 0;
}



sub run_goldengate_cmd() {
	my $self = shift;
	my ($cmd, $out) = @_;
	my $ret = 0;
	if ($^O =~ m/MSWin32/i ) {
		#my $Pword = "vbs\@123";
		my $cmd_line = "clientimp.exe $self->{'gg_bin'} \"$cmd\" $self->{'user'} $self->{'domain'} $self->{'password'}";
		$out = `$cmd_line`;
		VCSAG_LOGDBG_MSG(1, "GoldenGate command $cmd_line complete with Output : $out", 46, "$cmd", "$out");
	} else {
		$out = `$SU $self->{'owner'} -c "$LD_ENV; echo LD_LIBRARY_PATH - \\\$LD_LIBRARY_PATH; $self->{'gg_bin'}" << !
$cmd
!`;
	VCSAG_LOGDBG_MSG(1, "GoldenGate command $cmd complete with Output : $out", 12, "$cmd", "$out");
	}
	
	if ($out =~ /Error/i) {
                VCSAG_LOG_MSG("E", "Error in running Golden gate command $cmd :: $out", 13, "$cmd", "$out");
		$ret = 1;
        }
	return ($ret, $out);
}

#
# Function : check_manager_state 
# Desc     : Checks for golden gate manager state. if not running give try to start it 
#
sub check_manager_state() {
	my $self = shift;
	if ($^O =~ m/MSWin32/i ) {
		VCSAG_LOGDBG_MSG(1, "On windows OS, manager can be configured as service and set as auto start.", 48);
		return 0;
	}

	my $pid=`pgrep -x \"mgr\"`; #TODO: can we use golden gate cli instead of pgrep.
	my $retCode =  $?;
	if(!$retCode) {
		chomp $pid;
		VCSAG_LOG_MSG("I", "mgr process Id: $pid ", 49, "$pid");
		return 0;
	}

	VCSAG_LOG_MSG("W", "mgr process is not running. Give a try to start it. ", 52, "$pid");
	my $cmd = "start manager";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);
	if ($ret != 0) {
		VCSAG_LOG_MSG("W", "Failed to start manager.", 51);
		return $ret; #Don't fault the resource even if manager is not started. Either EXTRACAT or REPLICAT will take care for resource state.  
	}
	VCSAG_LOG_MSG("I", "Manager started succesfully.", 50);
	return $ret;
}

sub get_extract_state() {
        my $self = shift;
 	my $extract = shift;
        my $cmd;

        if ($extract eq "") {
                $extract = $self->{'extract'};
        }
        $cmd = "INFO EXTRACT $extract";       
        my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }

        my @out = split('\n', $out);
        while (@out) {
                $line = shift @out;
                if ($line =~ /EXTRACT/) {
                        my (@status_line)  = split (" ", $line);
                        while (@status_line) {
                                $status = shift @status_line;
                                if ($status =~ /Status/) {
                                        $self->{'extract_status'} = shift @status_line;
                                        last;
                                }
                        }
                        if ($self->{'extract_status'} eq "") {
                                VCSAG_LOG_MSG("E", "Failed to get Extract $self->{'extract'} status.", 34, "$self->{'extract'}");
                                $ret = 1;
                        }
                        last;
                }
        }
        VCSAG_LOGDBG_MSG(1, "Extract $self->{'extract'} status is $self->{'extract_status'}", 
				35, "$self->{'extract'}", "$self->{'extract_status'}");
        return ($ret, $self->{'extract_status'});
}



sub get_replicat_state() {
        my $self = shift;
        my $cmd;
        my $status;

        $cmd = "INFO REPLICAT $self->{'replicat'}";
        my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }

        my @out = split('\n', $out);
        while (@out) {
                $line = shift @out;
                if ($line =~ /REPLICAT/) {
                        my (@status_line)  = split (" ", $line);
                        while (@status_line) {
                                $status = shift @status_line;
                                if ($status =~ /Status/) {
                                        $self->{'replicat_status'} = shift @status_line;
                                        last;
                                }
                        }
                        if ($self->{'replicat_status'} eq "") {
                                VCSAG_LOG_MSG("E", "Failed to get Replicat $self->{'replicat'} status.", 36, "$self->{'replicat'}");
                                $ret = 1;
                        }
                        last;
                }
        }
        VCSAG_LOGDBG_MSG(1, "Replicat $self->{'replicat'} status is $self->{'replicat_status'}", 
				37, "$self->{'replicat'}", "$self->{'replicat_status'}");
        return ($ret, $self->{'replicat_status'});
}



sub start_extract() {
	my $self = shift;
        my $cmd = "START EXTRACT $self->{'extract'}";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	VCSAG_LOG_MSG("I", "Extract $self->{'extract'} started succesfully.", 14, "$self->{'extract'}");
        return $ret;
}



sub stop_extract() {
	my $self = shift;
        my $cmd;

	#First verify there is no lag
	$cmd = "LAG EXTRACT $self->{'extract'}";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	
	#As there is no lag proceed to stop extract

	$cmd = "STOP EXTRACT $self->{'extract'}";
        ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	VCSAG_LOG_MSG("I", "Extract $self->{'extract'} stopped succesfully.", 15, "$self->{'extract'}");
	return $ret;
}



sub start_datapump() {
        my $self = shift;
        my $cmd = "START EXTRACT $self->{'data_pump'}";
        my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
        VCSAG_LOG_MSG("I", "Extract $self->{'data_pump'} started succesfully.", 16, "$self->{'data_pump'}");
        return $ret;
}



sub stop_datapump() {
	my $self = shift;
        my $cmd;

	#First verify there is no lag
	$cmd = "LAG EXTRACT $self->{'data_pump'}";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	
	#As there is no lag proceed to stop extract

	$cmd = "STOP EXTRACT $self->{'data_pump'}";
        ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	VCSAG_LOG_MSG("I", "Extract $self->{'data_pump'} stopped succesfully.", 17, "$self->{'data_pump'}");
	return $ret;
}



sub start_replicat() {
	my $self = shift;
        my $cmd = "START REPLICAT $self->{'replicat'}";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	VCSAG_LOG_MSG("I", "Replicat $self->{'replicat'} started succesfully.", 18, "$self->{'replicat'}");
        return $ret;
}


sub stop_replicat() {
	my $self = shift;
        my $cmd;

	#First verify there is no lag
	$cmd = "LAG REPLICAT $self->{'replicat'}";
	my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	
	#As there is no lag proceed to stop extract

	$cmd = "STOP REPLICAT $self->{'replicat'}";
        ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	VCSAG_LOG_MSG("I", "Replicat $self->{'replicat'} stopped succesfully.", 19, "$self->{'replicat'}");
	return $ret;
}


sub switchover () {
	my $self = shift;
	my $cmd;

	#stop remote site extract first
	my $ret = $self->stop_extract_remote();

	if ($ret != 0) {
		return $ret;
	}
	

	#Remote site extract is stopped proceed with stopping replicat.
 	$ret = $self->stop_replicat();

        if ($ret != 0) {
                return $ret;
        }

	
	#Both remote extract and replicat has been stopped. Alter the local extract first.
	$cmd = "ALTER EXTRACT $self->{'extract'}, BEGIN NOW";
	($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
	
	#Now start the local extract and remote replicat
	$ret = $self->start_extract();

        if ($ret != 0) {
                return $ret;
        }

	$ret =  $self->start_replicat_remote();
	
	if ($ret != 0) {
		VCSAG_LOG_MSG("I", "Local Extract is in 'RUNNING' state. Remote Replicat failed to start.Start remote Replicat manually before retrying online operation.", 20);
                return $ret;
        }

	#start data pump if configured.
	if ($self->{'data_pump'} ne "") {
		$ret = $self->start_datapump();
	}
	return $ret;
}




sub failover () {
	my $self = shift;
        my $cmd;

	#remote site cluster is not RUNNING. stopped the replicat.
	my $ret = $self->stop_replicat();

        if ($ret != 0) {
                return $ret;
        }

	#Both remote extract and replicat has been stopped. Alter the local extract first.
        $cmd = "ALTER EXTRACT $self->{'extract'}, BEGIN NOW";
        ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }

        #Now start the local extract and remote replicat
        $ret = $self->start_extract();

        if ($ret != 0) {
                return $ret;
        }

	return $ret;
}

sub get_remote_replicat_state () {
	my $self = shift;
        my ($ret, $state);

        my ($ResID, $RemClus, $RemNode);
        $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
	if ($ret == 0) {
                VCSAG_LOG_MSG("I", "Executing GGStatus action entry point on Remote node $RemNode", 21, "$RemNode");

                my $out  = `$HARES -action $ResID GGStatus -sys $RemNode -clust $RemClus 2>&1`;
                $ret = $?;
		
		my @out = split('\n', $out);
	        while (@out) {
                	$line = shift @out;
                	if ($line =~ /Replicat/) {
                        	my (@status)  = split (" ", $line);
				$state = $status[3];
				last;
			}
		}
		
                if ($state eq "") {
                        #Remote action StopExtract Failed.
                        VCSAG_LOG_MSG("E", "Failed to get remote replicat state. output :: $out", 22, "$out");
                        return 1;
                }
                VCSAG_LOGDBG_MSG(1, "Remote action complete with return code : $ret Output : $out", 23, "$ret", "$out");
		$ret = 0;
        } else {
                VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 24);
		$ret = 1;
        }
        return ($ret, $state);
}

sub start_replicat_remote() {
	my $self = shift;
	my ($ResID, $RemClus, $RemNode) = @_;
        my $ret = 0;
        $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
        if ($ret == 0) {
                VCSAG_LOG_MSG("I", "Executing StartReplicat action entry point on Remote node $RemNode", 21, "$RemNode");

                my $out  = `$HARES -action $ResID StartReplicat -sys $RemNode -clust $RemClus`;
		$ret = $?;
                if ($ret != 0) {
                        #Remote action StopExtract Failed.
                        VCSAG_LOG_MSG("E", "Failed to to start remote replicat. output :: $out", 22, "$out");
                        return $ret;
                }
		VCSAG_LOGDBG_MSG(1, "Remote action complete with return code : $ret Output : $out", 23, "$ret", "$out");
        } else {
                VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 24);
        }
        return $ret;
}


sub stop_extract_remote {
        my $self = shift;
        my $ret = 0;
        my ($ResID, $RemClus, $RemNode);
        $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
        if ($ret == 0) {
                VCSAG_LOG_MSG("I", "Executing StopExtract action entry point on Remote node $RemNode", 25, "$RemNode");

                $out  = `$HARES -action $ResID StopExtract -sys $RemNode -clust $RemClus 2>&1`;
                $ret = $?;
                VCSAG_LOGDBG_MSG(1, "Remote action complete with return code : $ret Output : $out", 38, "$ret", "$out");
                if ($ret == 0) {
                        return $ret;
                }
                my $action_retcode = $self->get_action_retcode($out);
                if ($action_retcode == 1) {
                        #Remote action StopExtract Failed but Oracle database is stopped or faulted.
                        #this may be fc cable pullout scenario. Proceed for failover

                        if ($self->{'auto_takeover'} == 1) {
                                VCSAG_LOG_MSG("I", "Remote Oracle database is failure has been detected. Since AutoTakeover is set to 1, Attempting to failover the Oracle GoldenGate resource $self->{'resname'}", 39, "$self->{'resname'}");
                                return 0;
                        } else {
                                VCSAG_LOG_MSG("E", "Remote Oracle database failure has been detected. Since AutoTakeover is not set to 1, VCS does not attempt to failover the Oracle GoldenGate resource $self->{'resname'}. Administrative intervention required.", 40, "$self->{'resname'}");
                                return 1;
                        }
                } elsif ($action_retcode == 2) {
                        #Remote action StopExtract Failed.
                        VCSAG_LOG_MSG("E", "Failed to stop remote Extract with output :: $out", 26, "$out");
                        return $ret;
                } else {
                        VCSAG_LOG_MSG(E, "Remote action is completed with invalid return code : $action_retcode", 27, "$action_retcode");
                }
        } else {
		VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 28);
        }
        return $ret;
}



sub get_goldengate_state() {
	my $self = shift;
	my($Owner, $LD_ENV, $DBAMGR, $CONNECTSTR, $DGRole, $DGStatus, $OpenMode) = @_;
	my $cmd;
	my ($ret, $status);
	my (@lines, $i);

	$cmd = "INFO EXTRACT EXT1"; 
	($ret, $status) = $self->get_extract_state($self->{'extract'});
	
	
	VCSAG_LOGDBG_MSG(1, "GoldenGate Extract $self->{'extract'} return status : $ret Output : $status", 29, "$self->{'extract'}", "$ret", "$status");

	return ($ret, $status);
}

# Function : get_link_res_attr
# Desc     : Retrieve the instance attributes from Oracle resource
# Input    : $LinkRes - the resource name of the local DG link resource
# Input    : $NodeName - the resource name of the local DG link resource
# Output   : $Sid
# Output   : $Owner
# Output   : $Home
# Output   : $EnvFile - source file for env vars
# Returns  : 1 - the associated OraDGLink resource does not exist
#
sub get_link_res_attr {
	my $self = shift;
	my ($LinkRes, $NodeName, $Sid, $Owner, $Home, $EnvFile) = @_;
	
	my $i = 2;
	if ($^O =~ m/MSWin32/i ) {
		for my $ResAtt ("Sid","UserName","Domain","EncryptedPasswd")
		{	
			chomp($_[$i] = `$HARES -value $LinkRes $ResAtt $NodeName`);
			print "HARES -value $LinkRes $ResAtt $NodeName";
			$i += 1;
		} 
	} else {
		for my $ResAtt ("Sid","Owner","Home","EnvFile") 
		{	
			chomp($_[$i] = `$HARES -value $LinkRes $ResAtt $NodeName`);
			print "HARES -value $LinkRes $ResAtt $NodeName";
			$i += 1;
		}
	}
	return 1 if $_[2] eq "" || $_[3] eq "" || $_[4] eq "";
	return 0;
}


# Function : get_local_nodename
# Desc     : Retrieve the name of the local node
# Input    : None
# Output   : None
# Returns  : Name of the local node

sub get_local_nodename {
	my $self = shift;
	my $SysName = "${VCSCONF}/conf/sysname";
	my $GlobZoneName = "$VCSCONF/.vcshost";
	my $LocSys;

	if ( -r "$SysName" ) {
		open(SNAME, "<$SysName");
	} elsif ( -r "$GlobZoneName" ) {
		open(SNAME, "<$GlobZoneName");
	} else {
		open(SNAME, "uname -n |");
	}
	chomp($LocSys = <SNAME>);
	close SNAME;
	return $LocSys;
}

# Function : get_remote_info
# Desc     : Retrieve the name of the remote node and cluster
# Input    : $ResID
# Output   : $RemClust
# Output   : $RemNode
# Returns  : Name of the local node

sub get_remote_info {
        my $self = shift;
        my ($ResID, $RemClus, $RemNode) = @_;
        $ResID = $self->{'resname'};
        open(RDSP, "$HARES -display $ResID |");
        if ($self->is_global_group($ResID)) {
                while (<RDSP>) {
                        next unless /^$ResID\s+State\s/;
                        next if /\slocalclus/;
                        my @line = split();
                        ($RemClus, $RemNode) = split(':', $line[2]);
			my $out = `$HARES -action $ResID GGStatus -sys $RemNode -clust $RemClus 2>&1`;
			my $ret = $?;	
			last if $ret == 0;
			$RemClus = $RemNode = "";
                }
        } else {
                while (<RDSP>) {
                        next unless /^$ResID\s+State\s/;
                        next if /\sOFFLINE/;
                        my @line = split();
                        $RemNode = $line[3];
                        $RemClus = $RemNode = "";
                }
        }
        # Make sure the pipe is empty
        while (<RDSP>) {next;}
        close RDSP;
        VCSAG_LOGDBG_MSG(1, "Remote cluster : $RemClus and Remote node : $RemNode", 31, "$RemClus", "$RemNode");
        return 1 if ( $RemClus eq "" && $RemNode eq "" );
        $_[1] = $RemClus;
        $_[2] = $RemNode;
        return 0;
}

# Function : is_global_group
# Desc     : check if configured group is global group
# Input    : $ResID
# Returns  : 0 - not a global group(RDC)
# Returns  : 1 - global group

#check if configured group is global group
sub is_global_group {
        my $self = shift;
        my ($ResID) =  @_;
        my $ret = 0;
         my $cmd = "$HARES -value $ResID Group";
        my $Grp = `$cmd`;
        chomp $Grp;
        $cmd = "$HAGRP -value $Grp  ClusterList";
        my $ClusterList = `$cmd`;
        chomp $ClusterList;
        #$cmd = "hagrp -value $Grp  ClusterList";
        #my $ClusterList = `$cmd`;

        if ( $ClusterList eq "" ) {
                VCSAG_LOG_MSG("I", "GoldenGate Service Group $Grp is not configured as global group.", 32, "$Grp");
                $ret = 0;
        } else {
                VCSAG_LOG_MSG("I", "GoldenGate Service Group $Grp is configured as global group.", 33, "$Grp");
                $ret = 1;
        }
        return $ret;
}

# Function : set_instance_env
# Desc     : Checks the EnvFile, if not valide sets $ValidFile to 1
# Desc     : Else sets the Oracle related env vars 
# Input    : $EnvFile - source file for env vars
# Input    : $Owner 
# Input    : $Home 
# Output   : $ValidFile - 
# Returns  :  $LD_ENV - Env vars in format to be set 
# Note     :  This function sets LD_LIBRARY_PATH ( and equivalents) 
#          :  TWO_TASK, extra env vars 
#
sub set_instance_env {
	my $self = shift;
	my ($EnvFile, $Owner, $Home, $ValidFile) = @_;
	my $UseCSH = 0;
	my ($LD_ENV);
	my $EnvFileExists = 0;

	open (PASSWD, $PASSWDFILE);
	while (<PASSWD>) {
		next unless /^$Owner:/;
		$UseCSH = 1 if /csh$/;
		last;
	}
	# Make sure the pipe is empty
	while (<PASSWD>) {next;}
	close PASSWD;

	my $EXTRA_ENV= ggconfig::set_extra_env($UseCSH, $Home);
	if ( $EnvFile ) {
		
		#
		# It is the user's responsibility to ensure that EnvFile can
		# be sourced successfully using the login shell of $Usere 
		#
		# First ensure that the file is a Plain file AND a Text file.
		# Other file types (e.g. Named pipe) etc. cause $EnvFile to be
		# NOT sourced, with a message in the engine log saying so.
		#
		if ( -f $EnvFile && -T $EnvFile ) {
			#
			# Doing a 'test -r' on the file as user '$Owner' ensures that the 
			# permission are correct for user '$Owner'.
			# Execute permissions are NOT needed for $EnvFile to source it.
			#
			# Notes:
			#   * `` always invokes sh.
			#
			my $readerr = 0;
			$readerr = `$SU $Owner -c "test -r $EnvFile; echo $?" 2>&1`;
			if ( $readerr ) {
				$EnvFileExists = 1;	# read permission exists
				$_[3] = 0;		#is a valid file
			} else {
				$EnvFileExists = 0;	# error encountered by test -r
				$_[3]=1; 		# Not a valid file
			}
		} else {
			$EnvFileExists = 0;		# not a text file
			$_[3]=1; 			# Not a valid file
		}
	}
	# Important points from the man page of 'su':
	#   * Regarding LD_LIBRARY_PATH,
	#       Variables with LD_ prefix are removed for security  reasons.
	#       Thus,  su  bin will not retain previously exported variables
	#       with LD_ prefix while becoming user bin.
	#   * The environment  of  user root is passed along to user oracle,
	#       with the exception of $PATH, which is controlled by PATH and
	#       SUPATH in /etc/default/su.
	#   * $LANG environment variable: The su should work even if LANG=ja:
	#       If any  of  the  LC_*  variables  (  LC_CTYPE,  LC_MESSAGES,
	#       LC_TIME,   LC_COLLATE,  LC_NUMERIC,  and  LC_MONETARY)  (see
	#       environ(5)) are not set in the environment, the  operational
	#       behavior  of  su  for  each corresponding locale category is
	#       determined by the value of the LANG environment variable. If
	#       LC_ALL  is  set,  its contents are used to override both the
	#       LANG and the other LC_* variables.  If  none  of  the  above
	#       variables  are  set in the environment, the "C" (U.S. style)
	#       locale determines how su behaves.

	# NOTE: that the LD_LIBRARY_PATH should be set AFTER the su to $Owner,
	#	and NOT for environment of the user (root) running this script.
	#
	#	Be careful with the backslashes in this script.
	#
	my ($ldlibpath);
	if ( -d "$Home/lib64" ) {
		 $ldlibpath="/usr/lib:$Home/lib64:$self->{'goldengate_home'}:\\\$LD_LIBRARY_PATH";
	} else {
		$ldlibpath="/usr/lib:$Home/lib:$self->{'goldengate_home'}:\\\$LD_LIBRARY_PATH";
	}
	
	# If CSH  then setenv else export environment variables 
	if ($UseCSH) {
		$LD_ENV = "setenv LD_LIBRARY_PATH \\\"\\\"; ". $EXTRA_ENV;
		#
		# Note the .= concatenation operator below
		#
		if ($EnvFileExists) {
			$LD_ENV .= "source $EnvFile; ";
		}
		$LD_ENV .= "setenv LD_LIBRARY_PATH $ldlibpath;unsetenv TWO_TASK";
	} else {
		$LD_ENV = "LD_LIBRARY_PATH=\"\"; export LD_LIBRARY_PATH; ". $EXTRA_ENV;
		#
		# Note the .= concatenation operator below
		#
		if ($EnvFileExists) {
			$LD_ENV .= ". $EnvFile; ";
		}
		$LD_ENV .= "LD_LIBRARY_PATH=$ldlibpath; export LD_LIBRARY_PATH;TWO_TASK=; export TWO_TASK";
	}
	return $LD_ENV;
}






# Function : get_oracle_version
# Desc     : It is used to find the version of Oracle.
# Input    : $version - passed as reference. Oracle version will be copied.
#          : $release - passed as reference. Oracle release will be copied.
# Returns  : always 1
#
sub get_oracle_version
{
        my ($version, $release) = @_;
        my ($out, @out1);
        my ($lineval, $string, $str);

        my $Home = $ENV{'ORACLE_HOME'};
        if ($^O =~ m/MSWin32/i ) {
		$out = `$Home\\bin\\sqlplus -V`;
	  } else {
		$out = `$Home/bin/sqlplus -V`;
	  }
        @out1 = split(/\n/,$out);
        foreach $lineval (@out1) {
                chomp $lineval;
                if($lineval=~/Release/) {
                        $string = $lineval;
                        ($str = $string) =~ s/.*Release\s+(\d+)\.(\d+).*/$1/;
                        $release = $2;
                        last;
                }
        }

        $_[0] = $str;
        $_[1] = $release;

        return 1;
}


#check if Orcle resource is ONLINE.
sub is_oracle_running () {
        my $self = shift;
        my $ret = 1;
        my $Node = $self->get_local_nodename();

        open(RDSP1, "$HARES -state $self->{'ora_res'} |");
        while (<RDSP1>) {
                my @line = split();
                next if ($Node ne $line[2]);
                if ($line[3] eq "ONLINE") {
                        $ret = 0;
                }
                last;
        }
        # Make sure the pipe is empty
        while (<RDSP1>) {next;}
        close RDSP1;
        return $ret;
}


#Extract action entry point output and get the return code.`:w

sub get_action_retcode () {
        my $self = shift;
        my $out = shift;

        my $ret = 1;
        my $return = -1;

        #VCS WARNING V-16-13326 Resource (O_goldengate) - action (StartExtract) failed with return value (1)
        if ($out =~ /failed/) {
                my @retcode = split(' ', $out);
                while (@retcode) {
                        $ret = shift @retcode;
                        next unless ($ret eq "value");
                        $ret = shift @retcode;
                        my @retcode = split (/[()]/, $ret);
                        $return = $retcode[1];
                        last;
                }
        }
        if (($return ==  -1) or ($return == "")) {
                VCSAG_LOG_MSG("E", "Failed to get action return code from action output : $out.", 41, "$out");
        }
	VCSAG_LOGDBG_MSG(1, "Action returned return code :: $return", 42, "$return");
        return $return;
}

#Caluculate RPO. Agent shows Remote Extract lag and Local Replicat lag as the RPO.
sub get_current_rpo() {
	my $self = shift;
        my $out;
        my $ret = 0;
	my $ret1 = 0;

	#get extract lag from remote node
	$ret = $self->get_extract_lag_remote();
	
	#get replicat lag
	$ret =  $self->get_replicat_lag();
	
	return $ret;
}

#Get Extract Lag. This is used for RPO calculation
sub get_extract_lag() {
        my $self = shift;
        my $cmd;
        my $extract_lag;

        $cmd = "LAG EXTRACT $self->{'extract'}";
        my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
        my @out = split('\n', $out);
        while (@out) {
                $line = shift @out;
                if ($line =~ /EXTRACT/) {
                        last;
                }
        }
        pop (@out);
        pop (@out);
        $extract_lag = join("\n", @out);
        $extract_lag = "Extract $self->{'extract'}:\n".$extract_lag;
	if ($self->{'data_pump'} ne "") {
                $cmd = "LAG EXTRACT $self->{'data_pump'}";
                my ($ret, $out) = $self->run_goldengate_cmd($cmd);

                if ($ret != 0) {
                        return $ret;
                }
                my @out = split('\n', $out);
                while (@out) {
                        $line = shift @out;
                        if ($line =~ /EXTRACT/) {
                                last;
                        }
                }
                pop (@out);
                pop (@out);
                $datapump_lag = join("\n", @out);
                $datapump_lag = "DataPump $self->{'data_pump'}:\n".$datapump_lag;
                $extract_lag = $extract_lag."\n\n".$datapump_lag;
        }
	VCSAG_LOG_MSG("I", "$extract_lag", 44, "$extract_lag");
	print $extract_lag;
        return 0;
}

#Get Extract Lag from Remote node. This is used for RPO calculation
sub get_extract_lag_remote() {
        my $self = shift;
        my ($ResID, $RemClus, $RemNode) = @_;
        my $ret = 0;
	my $out = "";
        $ResID = $self->{'resname'};
        $ret = $self->get_remote_info($ResID, $RemClus, $RemNode);
        if ($ret == 0) {
                VCSAG_LOGDBG_MSG(1, "Executing GetExtractLag action entry point on Remote node $RemNode", 21, "$RemNode");

		$out  = `$HARES -action $ResID GetExtractLag -sys $RemNode -clust $RemClus 2>&1`;
                $ret = $?;
                if ($ret != 0) {
                        #Remote action GetExtractLag Failed.
                        VCSAG_LOG_MSG("E", "Failed to get Remote Extract Lag. output :: $out", 22, "$out");
                        return $ret;
                }
                VCSAG_LOGDBG_MSG(1, "Remote action complete with return code : $ret Output : $out", 23, "$ret", "$out");
        } else {
                VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 24);
		return 1;
        }
	my @extract_lag = split('\n', $out);
        shift @extract_lag;
        my $extract_lag =  join("\n", @extract_lag);
	VCSAG_LOG_MSG("I", "$extract_lag", 44, "$extract_lag");
	printf "$extract_lag\n\n";
	return 0;
}


#Get Replicat Lag. This is used for RPO calculation
sub get_replicat_lag() {
        my $self = shift;
        my $cmd;
        my $replicat_lag;

        $cmd = "LAG REPLICAT $self->{'replicat'}";
        my ($ret, $out) = $self->run_goldengate_cmd($cmd);

        if ($ret != 0) {
                return $ret;
        }
        my @out = split('\n', $out);
        while (@out) {
                $line = shift @out;
                if ($line =~ /REPLICAT/) {
                        last;
                }
        }
	pop (@out);
	pop (@out);
	pop (@out);

        $replicat_lag = join("\n", @out); 
	$replicat_lag = "Replicat $self->{'replicat'}:\n".$replicat_lag;
	VCSAG_LOG_MSG("I", "$replicat_lag", 45, "$replicat_lag");
	print "$replicat_lag";
	return 0;
}

#Calculated remaining OfflineTimeout value
sub get_offline_time(){
        my $self = shift;
        my $start_time = $self->{'start_time'};
        my $cur_time = time();

        if ($self->{'offline_timeout'} == 0 ){
                my $out = `$ag_i18n_inc::HATYPE -value GoldenGate OfflineTimeout`;
                $self->{'offline_timeout'} = $out;
        }

        my $remaining_time = $self->{'offline_timeout'} - ($cur_time - $self->{'start_time'});
        return $remaining_time;
}

1;
